(* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontlink.ml,v 1.5 2002/08/12 12:42:12 malekith Exp $
 *
 * Module implementation: Gontlink
 * Description: Gont linker.
 *)

open Gontutil
open Gontmodule
open Gontcodegen

open Unix

let check_order mi_lst =
  let check_module gmi =
    let rec find lst seen_self looking_for =
      match lst with
      | ({mi_name = n} as mi)::xs when (n = looking_for) ->
          if (mi.mi_init_fnc <> "" || mi.mi_fini_fnc <> "") &&
             seen_self then
               raise (Error (xf "module %s needs proper initialization \
                                 of %s, before it is used" 
                                 gmi.mi_name looking_for))
          else
               ()
      | {mi_name = n}::xs when (n = gmi.mi_name) ->
          find xs true looking_for
      | _::xs ->
          find xs seen_self looking_for
      | [] -> () in
    List.iter (fun n -> find mi_lst false n) gmi.mi_used_modules in
  List.iter check_module mi_lst

let make_archive target ldopts files =
  let mi_lst = List.map module_info_of_file files in
  let () = check_order mi_lst in
  let out_ch = open_out target in
  let () = save_module_info_list out_ch ldopts mi_lst in
  close_out out_ch

let err s = raise (Error s)

let split_filename x =
   let base = 
     try
       Filename.chop_extension x
     with Invalid_argument _ -> 
       err (xf "filename `%s' does not contain dot" x) in
   let ext = String.sub x (String.length base) 
                          (String.length x - String.length base) in
   (base, ext)


type link_info = 
  {
    li_target_prefix: string;
    li_file_list: (bool * string) list;
    li_search_path: string list;
    li_aux_ksi_file: string;
    li_init_name: string;
    li_fini_name: string;
    li_main_name: string;
    li_ldopts: string;
  }

let void_fnc n = 
  Ksi.FuncProto([Ksi.Extern], Ksi.Void, (ksi_name n), []) 
let commas = String.concat ", "
let protos li = [
  Ksi.TopComment("auxiliary linker file");
  Ksi.TopComment("following files were linked: " ^ 
                (commas (List.map (fun (g, n) -> n ^ 
                                        if g then "" else ":lib") 
                                  li.li_file_list)));
  void_fnc "gont_init_runtime";
  void_fnc "gont_run_exit_routines";
  void_fnc "gont_exit";
  Ksi.FuncProto([Ksi.Extern], Ksi.Void, 
                ksi_name "gont_register_exit_routine",
                [Ksi.FuncPtr(Ksi.Void, [Ksi.Ptr(Ksi.Void)]);
                 Ksi.Ptr(Ksi.Void)]);
  Ksi.FuncProto([Ksi.Extern], Ksi.Void, ksi_name "gont_set_argv",
                [Ksi.ptroff (); Ksi.Ptr(Ksi.Void)])
]

let output_aux_ksi_file li mi_lst =
  let out_ch = open_out li.li_aux_ksi_file in
  let ksiout = List.iter (Ksi.output_decl out_ch) in
  Ksi.read_target_conf li.li_target_prefix;
  Ksi.output_header out_ch;
  let make_go_linked {mi_name = n} =
    Ksi.Global([Ksi.Public], 
               Ksi.ptroff(), 
               ksi_name (n ^ "$$__go_file_linked__$")) in
  let out_proto n =
    if n <> "" then
      ksiout [void_fnc n] in
  let void_call n = 
    if n = "" then Ksi.Nop else Ksi.Call(ksi_name n, []) in
  let exns_init ini =
    let h : (string, unit) Hashtbl.t = Hashtbl.create 16 in
    let _ = List.iter (fun {mi_used_exns = e} -> 
                         List.iter (fun e -> 
                                        if not (Hashtbl.mem h e) then 
                                             Hashtbl.add h e ()) e) mi_lst in
    let output_exn e _ (no, init) =
      let n = ksi_name e in
      ksiout [Ksi.Global([Ksi.Public], Ksi.ptroff(), n)];
      (no + 1, (Ksi.Modify(Ksi.Ref(n), 
                           Ksi.IntConst(Ksi.ptroff(), string_of_int no))) ::
               init) in
    let (_, init) = Hashtbl.fold output_exn h (1, ini) in
    init in
  ksiout (protos li);
  ksiout (List.map make_go_linked mi_lst);
  
  List.iter (fun {mi_init_fnc = i; mi_fini_fnc = f} -> 
                                out_proto i; out_proto f) mi_lst;

  let fini_calls =
    List.rev (List.map (fun {mi_fini_fnc = i} -> void_call i) mi_lst) in
  let fini_body = Ksi.Bind(Ksi.CompoundList (fini_calls)) in
  ksiout [Ksi.Func([Ksi.Local], Ksi.Void,
                   ksi_name "gont_fini_modules", 
                   [Ksi.Ptr(Ksi.Void), ksi_name "dummy"], 
                   fini_body)];
  
  let init_calls =
    List.map (fun {mi_init_fnc = i} -> void_call i) mi_lst in
  let init_calls = (void_call "gont_init_runtime") :: init_calls in
  let register_fini = 
        Ksi.Call(ksi_name "gont_register_exit_routine", 
                 [Ksi.FuncRef(ksi_name "gont_fini_modules");
                  Ksi.Null(Ksi.Ptr(Ksi.Void))]) in
  let init_all = (exns_init []) @ [register_fini] @ init_calls in
  let init_body = 
    Ksi.Bind(Ksi.CompoundList (init_all)) in
  ksiout [Ksi.Func([Ksi.Public], Ksi.Void,
                   ksi_name li.li_init_name, [], init_body)];
                  
  begin
    if li.li_main_name <> "" then
      ksiout [
        Ksi.Func([Ksi.Public], Ksi.Void, ksi_name li.li_main_name, 
           [Ksi.c_type Ksi.C_int, ksi_name "argc";
            Ksi.c_type Ksi.C_ptr, ksi_name "argv"],
           Ksi.Bind(Ksi.CompoundList [
               Ksi.Call(ksi_name "gont_set_argv", 
                              [Ksi.Ref(ksi_name "argc");
                               Ksi.Ref(ksi_name "argv")]);
               void_call li.li_init_name;
               void_call "gont_run_exit_routines";
               void_call "gont_exit"]))]
  end;

  Ksi.output_trailer out_ch;
  close_out out_ch

type module_desc =
  {
    md_mi: module_info;
    mutable md_needed: bool;
    mutable md_been_here: bool;
  }
  
let link li =
  let ldopts = ref li.li_ldopts in
  let find_file name =
    let file_exists_in p =
      let n = Filename.concat p name in
      try
        let _ = Unix.stat n in
        true
      with Unix_error (ENOENT, _, _) -> false in
    try
      Filename.concat (List.find file_exists_in li.li_search_path) name
    with Not_found ->
      err (xf "cannot find file `%s' in search path" name) in
  let rec read_mi res lnk_res lst =
    match lst with
    | (greedy, x)::xs ->
       let (base, ext) = split_filename x in
       let (added, lnk_added) = 
         match ext with
         | ".go" ->
           ([{md_mi = module_info_of_file x; 
              md_needed = greedy;
              md_been_here = false}], 
            (base ^ ".o"))
         | ".ga" ->
           let name = find_file x in
           let (base, ext) = split_filename name in
           let (ldopt, mi_list) = module_info_list_of_file name in
           let _ = ldopts := !ldopts ^ ldopt in
           let md_list = List.map (fun mi -> {md_mi = mi; 
                                              md_needed = greedy;
                                              md_been_here = false}) mi_list in
           let (dir, p) = (Filename.dirname base, Filename.basename base) in
           (List.rev md_list, (dir ^ "/libgont" ^ p ^ ".a"))
         | ".o" -> ([], x)
         | _ -> err (xf "don't know what to do with `%s'" x) in
       read_mi (added @ res) (lnk_added::lnk_res) xs
    | [] -> (List.rev res, List.rev lnk_res) in
  let (md_lst, lnk_res) = read_mi [] [] li.li_file_list in
  let () = check_order (List.map (fun {md_mi = mi} -> mi) md_lst) in
  let fnd n =
    try
      List.find (fun {md_mi = {mi_name = k}} -> n = k) md_lst
    with Not_found -> err (xf "unresolved reference to module %s" n) in
  let rec mark_needed mi =
    match mi with
    | {md_been_here = true}
    | {md_needed = false} -> ()
    | x ->
        x.md_been_here <- true;
        List.iter (fun n -> 
                    let md = fnd n in
                    md.md_needed <- true;
                    mark_needed md) x.md_mi.mi_used_modules in
  let () = List.iter mark_needed md_lst in
  let mi_lst = List.map (fun {md_mi = mi} -> mi) 
                        (List.filter (fun {md_needed = n} -> n) md_lst) in
  let () = check_order mi_lst in
  let () = output_aux_ksi_file li mi_lst in
  (!ldopts, List.rev lnk_res)
;;

