/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: hashtbl.g,v 1.1 2002/08/12 12:42:21 malekith Exp $
 *
 * Standard library for Gont, Hashtbl module implementation.
 *
 * Mutable hash table.
 */

open Hashtbl;

struct <'a>entry {
	<*[string, 'a]>List::t lst;
};

struct <'a>t {
	int size;
	int nelem;
	<<'a>entry>Array::t tab;
};

<'a>t create(int sz)
{
	let ini = (fun <'a>entry (int x) (({lst = List::Nil}))) in
	return ({ nelem = 0,
		  size = sz,
		  tab = Array::init(sz, ini) });
}

void iter(<'a>t h, *(void (string, 'a)) f)
{
	void f2(*[string, 'a] p)
	{
		let [k, v] = p in
		f(k, v);
	}
	
	for (int i = 0; i < h.size; i++)
		List::iter(f2, (Array::get(h.tab, i)).lst);
}

void check_rehash(<'a>t h)
{
	if (h.nelem < h.size)
		return;
		
	<'a>t n = create(h.nelem * 3);

	let f = (fun void (string k, 'a v) (add(n, k, v))) in
	iter(h, f);

	h.tab = n.tab;
	h.size = n.size;
}

'a find_lst(<*[string, 'a]>List::t lst, string key)
{
	switch lst {
	case List::Nil:
		raise Not_found;
	case List::Cons[[k, v], tl]:
		if (k == key)
			return v;
		else
			return find_lst(tl, key);
	}
}

'a find(<'a>t h, string k)
{
	let e = Array::get(h.tab, String::hash(k) % h.size) in
	return find_lst(e.lst, k);
}

<'a>List::t find_all_lst(<*[string, 'a]>List::t lst, string key)
{
	switch lst {
	case List::Nil:
		return List::Nil;
	case List::Cons[[k, v], tl]:
		if (k == key)
			return List::Cons[v, find_all_lst(tl, key)];
		else
			return find_all_lst(tl, key);
	}
}

<'a>List::t find_all(<'a>t h, string k)
{
	let e = Array::get(h.tab, String::hash(k) % h.size) in
	return find_all_lst(e.lst, k);
}

bool mem(<'a>t h, string k)
{
	bool xfind(<*[string, 'a]>List::t lst)
	{
		switch lst {
		case List::Nil:
			return false;
		case List::Cons[[key, _], tl]:
			if (k == key)
				return true;
			else
				return xfind(tl);
		}
	}
	
	let e = Array::get(h.tab, String::hash(k) % h.size) in
	return xfind(e.lst);
}

void add(<'a>t h, string k, 'a v)
{
	h.nelem++;
	check_rehash(h);
	let e = Array::get(h.tab, String::hash(k) % h.size) in
	e.lst = List::Cons[[k, v], e.lst];
}
